from aioredis import Redis

from cenao.app import AppFeature


class RedisStorageAppFeature(AppFeature):
    NAME = 'redis_storage'

    redis: Redis

    def __init__(self, redis: str = 'redis', name=None):
        super().__init__(name)
        self._redis = redis

    def on_init(self):
        self.redis = self.app.ft.get(self._redis)



from cenao.app import AppFeature
from cenao.view import View


class HealthView(View):
    ROUTE = '/health'

    async def get(self):
        for ft in self.app.ft:
            try:
                await ft.check_health()
            except Exception as e:
                self.logger.exception('Health check failed', exc_info=e)
                return {'ok': False}

        return {'ok': True}


class HealthAppFeature(AppFeature):
    NAME = 'health'
    VIEWS = [HealthView]



import logging
from typing import TYPE_CHECKING, Dict, Union, Generator, Any

from aiohttp import abc, hdrs, web
from aiohttp.web_exceptions import HTTPMethodNotAllowed

from cenao.errors import ViewError, ServerError

if TYPE_CHECKING:
    from cenao.app import Application, AppFeature


class View(abc.AbstractView):
    ROUTE: str

    logger: logging.Logger
    ft: 'AppFeature'

    @property
    def app(self) -> 'Application':
        return self.ft.app

    @classmethod
    def init(cls, ft: 'AppFeature'):
        cls.ft = ft
        cls.logger = logging.getLogger(cls.__name__)

    @property
    def remote_ip(self) -> str:
        if xff := self.request.headers.get('X-FORWARDED-FOR', ''):
            return xff
        if xri := self.request.headers.get('X-REAL-IP', ''):
            return xri
        return self.request.remote

    async def _iter(self) -> web.StreamResponse:
        if self.request.method not in hdrs.METH_ALL:
            self._raise_allowed_methods()
        method = getattr(self, self.request.method.lower(), None)
        if method is None:
            self._raise_allowed_methods()

        try:
            resp = await method()
        except ViewError as ve:
            self.logger.warning('Got an error while handling request: %r', ve)
            return web.json_response({
                'ok': False,
                'code': ve.code,
                'error': ve.error_msg(),
                'reason': ve.reason(),
            }, status=ve.status)
        except Exception as e:
            self.logger.warning('Got an exception while handling request: %r', e)
            e = ServerError()
            return web.json_response({
                'ok': False,
                'code': e.code,
                'error': e.error_msg(),
                'reason': e.reason(),
            }, status=e.status)
        if resp is None:
            return web.json_response({'ok': True})

        if isinstance(resp, dict):
            return web.json_response({'ok': True, 'result': resp})

        return resp

    def __await__(self) -> Generator[Any, None, abc.StreamResponse]:
        return self._iter().__await__()

    def _raise_allowed_methods(self) -> None:
        allowed_methods = {m for m in hdrs.METH_ALL if hasattr(self, m.lower())}
        raise HTTPMethodNotAllowed(self.request.method, allowed_methods)

    async def get(self) -> Union[Dict, None, web.Response]:
        pass

    async def post(self) -> Union[Dict, None, web.Response]:
        pass

    async def put(self) -> Union[Dict, None, web.Response]:
        pass

    async def delete(self) -> Union[Dict, None, web.Response]:
        pass



from typing import Any, Dict, Optional

import aiohttp
import datetime
import hashlib
import hmac
import urllib.parse

from cenao.app import AppFeature


class S3Client:
    _endpoint: str
    _region: str
    _bucket: str
    _access: str
    _secret: str
    _session: aiohttp.ClientSession

    def __init__(self, endpoint: str, region: str, bucket: str, access: str, secret: str):
        self._endpoint = endpoint
        self._region = region
        self._bucket = bucket
        self._access = access
        self._secret = secret

        self._session = aiohttp.ClientSession()

    async def upload_data(self, key: str, payload: bytes):
        path = f'/{self._bucket}/{key}'
        headers = self._sign(path=path, method='PUT', payload=payload)
        resp = await self._session.put(url=f'https://{self._endpoint}{path}', headers=headers, data=payload)
        resp.raise_for_status()

    def _sign(
        self,
        path: str,
        method: str,
        payload: Optional[bytes] = None,
        headers: Optional[Dict[str, Any]] = None,
        query: Optional[Dict[str, Any]] = None,
    ) -> Dict:
        if headers is None:
            headers = {}

        if query is None:
            query = {}

        now = datetime.datetime.utcnow()
        amzdate = now.strftime('%Y%m%dT%H%M%SZ')
        datestamp = now.strftime('%Y%m%d')
        payload_hash = hashlib.sha256(payload).hexdigest() if payload else None
        credential_scope = f'{datestamp}/{self._region}/s3/aws4_request'

        headers_lower = {
            header_key.lower(): ' '.join(header_value.split())
            for header_key, header_value in headers.items()
        }

        _headers = {
            'host': self._endpoint,
            'x-amz-date': amzdate,
            **headers_lower,
        }

        if payload_hash:
            _headers['x-amz-content-sha256'] = payload_hash

        header_keys = sorted(_headers.keys())
        signed_headers = ';'.join(header_keys)

        def signature():
            def canonical_request():
                canonical_uri = urllib.parse.quote(path, safe='/~')
                quoted_query = sorted(
                    (urllib.parse.quote(key, safe='~'), urllib.parse.quote(value, safe='~'))
                    for key, value in query.items()
                )
                canonical_querystring = '&'.join(f'{key}={value}' for key, value in quoted_query)
                canonical_headers = ''.join(f'{key}:{_headers[key]}\n' for key in header_keys)

                canonical_resp = f'{method}\n{canonical_uri}\n{canonical_querystring}\n' + \
                                 f'{canonical_headers}\n{signed_headers}'
                if payload_hash:
                    canonical_resp += f'\n{payload_hash}'
                return canonical_resp

            def sign(key, msg):
                return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

            string_to_sign = f'AWS4-HMAC-SHA256\n{amzdate}\n{credential_scope}\n' + \
                             hashlib.sha256(canonical_request().encode('utf-8')).hexdigest()

            date_key = sign(('AWS4' + self._secret).encode('utf-8'), datestamp)
            region_key = sign(date_key, self._region)
            service_key = sign(region_key, 's3')
            request_key = sign(service_key, 'aws4_request')
            return sign(request_key, string_to_sign).hex()
        resp = {
            **headers,
            'x-amz-date': amzdate,
            'Authorization': f'AWS4-HMAC-SHA256 Credential={self._access}/{credential_scope}, '
                             f'SignedHeaders={signed_headers}, Signature=' + signature(),
        }

        if payload_hash:
            resp['x-amz-content-sha256'] = payload_hash

        return resp

    async def close(self):
        if self._session:
            await self._session.close()


class S3AppFeature(AppFeature):
    NAME = 's3'

    client: S3Client

    async def on_startup(self):
        self.client = S3Client(
            endpoint=self.config.get('endpoint'),
            region=self.config.get('region', 'us-east-1'),
            bucket=self.config.get('bucket'),
            access=self.config.get('access'),
            secret=self.config.get('secret'),
        )

    async def on_shutdown(self):
        if self.client:
            await self.client.close()



class ViewError(Exception):
    code: int
    status: int = 500
    reason_format: str = ''

    @classmethod
    def name(cls) -> str:
        return cls.__name__[:-5]

    def __repr__(self):
        return '<{} code={} reason={!r}>'.format(
            self.__class__.__name__, self.code, self.reason()
        )

    def reason(self):
        return self.reason_format.format(*self.args)

    def error_msg(self):
        return '{:d}: {}'.format(self.code, self.reason())


class ClientError(ViewError):
    code = 4000
    status = 400
    reason_format = 'Unknown client error'


class ClientUnauthorizedError(ViewError):
    code = 4001
    status = 401
    reason_format = 'Unauthorized error'


class ClientForbiddenError(ViewError):
    code = 4003
    status = 403
    reason_format = 'Forbidden error'


class ClientMethodNotAllowedError(ViewError):
    code = 4005
    status = 405
    reason_format = 'Method not allowed error'


class ClientConflictError(ViewError):
    code = 4009
    status = 409
    reason_format = 'Conflict error'


class ServerError(ViewError):
    code = 5000
    status = 500
    reason_format = 'Unknown server error'



import asyncio
from typing import List

from aiohttp import web
from aiohttp.web_runner import TCPSite, AppRunner
from prometheus_client import generate_latest, Counter

from cenao.app import AppFeature
from cenao.view import View


class MetricsView(View):
    ROUTE = '/metrics'

    async def get(self):
        response = generate_latest().decode('utf-8')

        return web.Response(
            status=200,
            text=response,
        )


class WebAppFeature(AppFeature):
    NAME = 'web'

    VIEWS = [MetricsView]

    host: str
    port: int

    aiohttp_app: web.Application
    runner: AppRunner
    _ws: List[web.WebSocketResponse]

    PROMETHEUS_ROUTE_METRIC: Counter

    def on_init(self):
        self.host = self.config.get('host', '0.0.0.0')
        self.port = int(self.config.get('port', 8000))

        self._ws = []

        self.PROMETHEUS_ROUTE_METRIC = Counter(
            'route',
            documentation='Total route calls',
            labelnames=('path', 'method', 'status'),
            namespace=self.app.NAME,
        )

        @web.middleware
        async def prometheus_route_call_count(request: web.Request, handler) -> web.Response:
            resp: web.Response = await handler(request)
            self.PROMETHEUS_ROUTE_METRIC.labels(
                request.match_info.route.resource.canonical,
                request.method,
                resp.status
            ).inc()
            return resp

        self.aiohttp_app = web.Application(
            loop=self.app.loop,
            middlewares=[prometheus_route_call_count]
        )
        routes_len = 0
        for ft in self.app.ft:
            for view in ft.VIEWS:
                view.init(ft)
                self.aiohttp_app.router.add_view(view.ROUTE, view)
                routes_len += 1
        self.logger.info('Registered %d routes', routes_len)

        self.runner = AppRunner(self.aiohttp_app, logger=self.logger)

    async def on_startup(self):
        self.logger.info('Starting on %s:%d', self.host, self.port)
        await self.runner.setup()
        site = TCPSite(
            self.runner,
            self.host,
            self.port,
        )
        await site.start()

    async def on_shutdown(self):
        self.logger.info('Stopping webserver')

        for ws in self._ws:
            await ws.close()

        await self.runner.cleanup()

    def get_websocket_response(self) -> web.WebSocketResponse:
        ws = web.WebSocketResponse()
        self._ws.append(ws)
        return ws
